package test.xmpp.xmpploadtest;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.DefaultParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Option;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;
import org.jivesoftware.smack.SmackException;
import org.jivesoftware.smackx.muc.MultiUserChat;

import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.List;
import java.util.Properties;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.logging.Logger;

import static java.util.concurrent.TimeUnit.MINUTES;
import static java.util.concurrent.TimeUnit.SECONDS;

/**
 * An XMPP load tester in Java, based on {@code loadtest.c}.
 *
 * This program simulates multiple XMPP users, creates a connection to an XMPP
 * server, logs in users, requests the roster and declares user presence. Then,
 * the users will send periodic XMPP messages to chat rooms to create network
 * traffic and server load. Thus a load test of a connected "real" client is
 * possible as well.
 *
 * @author ikost
 */
public final class XmppLoadTest {

    private static final Logger LOG = Logger.getLogger(XmppLoadTest.class.getName());

    // ~~~~~~~~~~~~~~~~~~~ CONSTANTS ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    /**
     * Send messages/attachments periodically in another thread.
     */
    private static final ScheduledExecutorService scheduler = Executors.newScheduledThreadPool(1);
    /**
     * Properties file that contains the delays of the messages/attachments.
     */
    private static final String PROPERTIES_FILE = "loadxmpptest.properties";
    /**
     * Path of big attachment.
     */
    private static final String DEFAULT_BIG_FILE_NAME_PATH = "blob.txt";

    /**
     * Path of small attachment.
     */
    private static final String DEFAULT_SMALL_FILE_NAME_PATH = "test.txt";
    /**
     * Default XMPP server port.
     */
    public static final int DEFAULT_PORT = 5222;

    /**
     * Default password for all users.
     */
    private static final String DEFAULT_PASSWORD = "a";

    /**
     * Default number of users.
     */
    private static final int DEFAULT_NUM_OF_USERS = 49;

    /**
     * Default number of rooms.
     */
    private static final int DEFAULT_NUM_OF_ROOMS = 5;

    /**
     * Maximum number of users per session.
     */
    private static final int MAX_NUM_OF_USERS_PER_SESSION = 256;

    /**
     * Short chat message.
     */
    private static final String MESSAGE = "lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do eiusmod tempor incididunt ut labore et dolore magne aliqua.";

    /**
     * Long chat message.
     */
    private static final String LONG_MESSAGE = "Nam quis nulla. Integer malesuada. In in enim a arcu imperdiet malesuada. Sed vel lectus. Donec odio urna, tempus molestie, porttitor ut, iaculis quis, sem. Phasellus rhoncus. Aenean id metus id velit ullamcorper pulvinar. Vestibulum fermentum tortor id mi. Pellentesque ipsum. Nulla non arcu lacinia neque faucibus fringilla. Nulla non lectus sed nisl molestie malesuada. Proin in tellus sit amet nibh dignissim sagittis. Vivamus luctus egestas leo. Maecenas sollicitudin. Nullam rhoncus aliquam metus. Etiam egestas wisi a erat.Lorem ipsum dolor sit amet, consectetuer adipiscing elit. Nullam feugiat, turpis at pulvinar vulputate, erat libero tristique tellus, nec bibendum odio risus sit amet ante. Aliquam erat volutpat. Nunc auctor. Mauris pretium quam et urna. Fusce nibh. Duis risus. Curabitur sagittis hendrerit ante. Aliquam erat volutpat. Vestibulum erat nulla, ullamcorper nec, rutrum non, nonummy ac, erat. Duis condimentum augue id magna semper rutrum. Nullam justo enim, consectetuer nec, ullamcorper ac, vestibulum in, elit. Proin pede metus, vulputate nec, fermentum fringilla, vehicula vitae, justo. Fusce consectetuer risus a nunc. Aliquam ornare wisi eu metus. Integer pellentesque quam vel velit. Duis pulvinar.Lorem ipsum dolor sit amet, consectetuer adipiscing elit. Morbi gravida libero nec velit. Morbi scelerisque luctus velit. Etiam dui sem, fermentum vitae, sagittis id, malesuada in, quam. Proin mattis lacinia justo. Vestibulum facilisis auctor urna. Aliquam in lorem sit amet leo accumsan lacinia. Integer rutrum, orci vestibulum ullamcorper ultricies, lacus quam ultricies odio, vitae placerat pede sem sit amet enim. Phasellus et lorem id felis nonummy placerat. Fusce dui leo, imperdiet in, aliquam sit amet, feugiat eu, orci. Aenean vel massa quis mauris vehicula lacinia. Quisque tincidunt scelerisque libero. Maecenas libero. Etiam dictum tincidunt diam. Donec ipsum massa, ullamcorper in, auctor et, scelerisque sed, est. Suspendisse nisl. Sed convallis magna eu sem. Cras pede libero, dapibus nec, pretium sit amet, tempor quis, urna.Etiam posuere quam ac quam. Maecenas aliquet accumsan leo. Nullam dapibus fermentum ipsum. Etiam quis quam. Integer lacinia. Nulla est. Nulla turpis magna, cursus sit amet, suscipit a, interdum id, felis. Integer vulputate sem a nibh rutrum consequat. Maecenas lorem. Pellentesque pretium lectus id turpis. Etiam sapien elit, consequat eget, tristique non, venenatis quis, ante. Fusce wisi. Phasellus faucibus molestie nisl. Fusce eget urna. Curabitur vitae diam non enim vestibulum interdum. Nulla quis diam. Ut tempus purus at lorem.In sem justo, commodo ut, suscipit at, pharetra vitae, orci. Duis sapien nunc, commodo et, interdum suscipit, sollicitudin et, dolor. Pellentesque habitant morbi tristique senectus et netus et malesuada fames ac turpis egestas. Aliquam id dolor. Class aptent taciti sociosqu ad litora torquent per conubia nostra, per inceptos hymenaeos. Mauris dictum facilisis augue. Fusce tellus. Pellentesque arcu. Maecenas fermentum, sem in pharetra pellentesque, velit turpis volutpat ante, in pharetra metus odio a lectus. Sed elit dui, pellentesque a, faucibus vel, interdum nec, diam. Mauris dolor felis, sagittis at, luctus sed, aliquam non, tellus. Etiam ligula pede, sagittis quis, interdum ultricies, scelerisque eu, urna. Nullam at arcu a est sollicitudin euismod. Praesent dapibus. Duis bibendum, lectus ut viverra rhoncus, dolor nunc faucibus libero, eget facilisis enim ipsum id lacus. Nam sed tellus id magna elementum tincidunt.Morbi a metus. Phasellus enim erat, vestibulum vel, aliquam a, posuere eu, velit. Nullam sapien sem, ornare ac, nonummy non, lobortis a, enim. Nunc tincidunt ante vitae massa. Duis ante orci, molestie vitae, vehicula venenatis, tincidunt ac, pede. Nulla accumsan, elit sit amet varius semper, nulla mauris mollis quam, tempor suscipit diam nulla vel leo. Etiam commodo dui eget wisi. Donec iaculis gravida nulla. Donec quis nibh at felis congue commodo. Etiam bibendum elit eget erat.Praesent in mauris eu tortor porttitor accumsan. Mauris suscipit, ligula sit amet pharetra semper, nibh ante cursus purus, vel sagittis velit mauris vel metus. Aenean fermentum risus id tortor. Integer imperdiet lectus quis justo. Integer tempor. Vivamus ac urna vel leo pretium faucibus. Mauris elementum mauris vitae tortor. In dapibus augue non sapien. Aliquam ante. Curabitur bibendum justo non orci.Morbi leo mi, nonummy eget, tristique non, rhoncus non, leo. Nullam faucibus mi quis velit. Integer in sapien. Fusce tellus odio, dapibus id, fermentum quis, suscipit id, erat. Fusce aliquam vestibulum ipsum. Aliquam erat volutpat. Pellentesque sapien. Cras elementum. Nulla pulvinar eleifend sem. Cum sociis natoque penatibus et magnis dis parturient montes, nascetur ridiculus mus. Quisque porta. Vivamus porttitor turpis ac leo.Maecenas ipsum velit, consectetuer eu, lobortis ut, dictum at, dui. In rutrum. Sed ac dolor sit amet purus malesuada congue. In laoreet, magna id viverra tincidunt, sem odio bibendum justo, vel imperdiet sapien wisi sed libero. Suspendisse sagittis ultrices augue. Mauris metus. Nunc dapibus tortor vel mi dapibus sollicitudin. Etiam posuere lacus quis dolor. Praesent id justo in neque elementum ultrices. Class aptent taciti sociosqu ad litora torquent per conubia nostra, per inceptos hymenaeos. In convallis. Fusce suscipit libero eget elit. Praesent vitae arcu tempor neque lacinia pretium. Morbi imperdiet, mauris ac auctor dictum, nisl ligula egestas nulla, et sollicitudin sem purus in lacus.Aenean placerat. In vulputate urna eu arcu. Aliquam erat volutpat. Suspendisse potenti. Morbi mattis felis at nunc. Duis viverra diam non justo. In nisl. Nullam sit amet magna in magna gravida vehicula. Mauris tincidunt sem sed arcu. Nunc posuere. Nullam lectus justo, vulputate eget, mollis sed, tempor sed, magna. Cum sociis natoque penatibus et magnis dis parturient montes, nascetur ridiculus mus. Etiam neque. Curabitur ligula sapien, pulvinar a, vestibulum quis, facilisis vel, sapien. Nullam eget nisl. Donec vitae arcu.Nam quis nulla. Integer malesuada. In in enim a arcu imperdiet malesuada. Sed vel lectus. Donec odio urna, tempus molestie, porttitor ut, iaculis quis, sem. Phasellus rhoncus. Aenean id metus id velit ullamcorper pulvinar. Vestibulum fermentum tortor id mi. Pellentesque ipsum. Nulla non arcu lacinia neque faucibus fringilla. Nulla non lectus sed nisl molestie malesuada. Proin in tellus sit amet nibh dignissim sagittis. Vivamus luctus egestas leo. Maecenas sollicitudin. Nullam rhoncus aliquam metus. Etiam egestas wisi a erat.Lorem ipsum dolor sit amet, consectetuer adipiscing elit. Nullam feugiat, turpis at pulvinar vulputate, erat libero tristique tellus, nec bibendum odio risus sit amet ante. Aliquam erat volutpat. Nunc auctor. Mauris pretium quam et urna. Fusce nibh. Duis risus. Curabitur sagittis hendrerit ante. Aliquam erat volutpat. Vestibulum erat nulla, ullamcorper nec, rutrum non, nonummy ac, erat. Duis condimentum augue id magna semper rutrum. Nullam justo enim, consectetuer nec, ullamcorper ac, vestibulum in, elit. Proin pede metus, vulputate nec, fermentum fringilla, vehicula vitae, justo. Fusce consectetuer risus a nunc. Aliquam ornare wisi eu metus. Integer pellentesque quam vel velit. Duis pulvinar.Lorem ipsum dolor sit amet, consectetuer adipiscing elit. Morbi gravida libero nec velit. Morbi scelerisque luctus velit. Etiam dui sem, fermentum vitae, sagittis id, malesuada in, quam. Proin mattis lacinia justo. Vestibulum facilisis auctor urna. Aliquam in lorem sit amet leo accumsan lacinia. Integer rutrum, orci vestibulum ullamcorper ultricies, lacus quam ultricies odio, vitae placerat pede sem sit amet enim. Phasellus et lorem id felis nonummy placerat. Fusce dui leo, imperdiet in, aliquam sit amet, feugiat eu, orci. Aenean vel massa quis mauris vehicula lacinia. Quisque tincidunt scelerisque libero. Maecenas libero. Etiam dictum tincidunt diam. Donec ipsum massa, ullamcorper in, auctor et, scelerisque sed, est. Suspendisse nisl. Sed convallis magna eu sem. Cras pede libero, dapibus nec, pretium sit amet, tempor quis, urna.Etiam posuere quam ac quam. Maecenas aliquet accumsan leo. Nullam dapibus fermentum ipsum. Etiam quis quam. Integer lacinia. Nulla est. Nulla turpis magna, cursus sit amet, suscipit a, interdum id, felis. Integer vulputate sem a nibh rutrum consequat. Maecenas lorem. Pellentesque pretium lectus id turpis. Etiam sapien elit, consequat eget, tristique non, venenatis quis, ante. Fusce wisi. Phasellus faucibus molestie nisl. Fusce eget urna. Curabitur vitae diam non enim vestibulum interdum. Nulla quis diam. Ut tempus purus at lorem.In sem justo, commodo ut, suscipit at, pharetra vitae, orci. Duis sapien nunc, commodo et, interdum suscipit, sollicitudin et, dolor. Pellentesque habitant morbi tristique senectus et netus et malesuada fames ac turpis egestas. Aliquam id dolor. Class aptent taciti sociosqu ad litora torquent per conubia nostra, per inceptos hymenaeos. Mauris dictum facilisis augue. Fusce tellus. Pellentesque arcu. Maecenas fermentum, sem in pharetra pellentesque, velit turpis volutpat ante, in pharetra metus odio a lectus. Sed elit dui, pellentesque a, faucibus vel, interdum nec, diam. Mauris dolor felis, sagittis at, luctus sed, aliquam non, tellus. Etiam ligula pede, sagittis quis, interdum ultricies, scelerisque eu, urna. Nullam at arcu a est sollicitudin euismod. Praesent dapibus. Duis bibendum, lectus ut viverra rhoncus, dolor nunc faucibus libero, eget facilisis enim ipsum id lacus. Nam sed tellus id magna elementum tincidunt.Morbi a metus. Phasellus enim erat, vestibulum vel, aliquam a, posuere eu, velit. Nullam sapien sem, ornare ac, nonummy non, lobortis a, enim. Nunc tincidunt ante vitae massa. Duis ante orci, molestie vitae, vehicula venenatis, tincidunt ac, pede. Nulla accumsan, elit sit amet varius semper, nulla mauris mollis quam, tempor suscipit diam nulla ve nulla nulla nulla nulla nulla";

    /**
     * File content for small attachments in Base64 encoding.
     */
    private static final String FILE_CONTENT = "SGVsbG8gd29ybGQsIHdlIGhhdmUgc29tZSBjb250ZW50IQo=";  // Hello world, we have some content!

    private static final int DEFAULT_DELAY_TO_SEND_MESSAGES_MILLISECONDS = 1000; // 1000 millis or 1"
    private static final int DEFAULT_SHORT_MESSAGES_DELAY_SECONDS = 100;  // 100"
    private static final int DEFAULT_LONG_MESSAGES_DELAY_SECONDS = 60; // 1'
    private static final int SMALL_ATTACHMENTS_DELAY_MINUTES = 1; // 1'
    private static final int BIG_ATTACHMENTS_DELAY_MINUTES = 5;  // 5'
    private static final int MINUTE_IN_MILLISECONDS = 60 * 1000;

    // ~~~~~~~~~~~~~~~~~~~ PROPERTIES ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    private static int shortMessagesDelayInSeconds = DEFAULT_SHORT_MESSAGES_DELAY_SECONDS;
    private static int longMessagesDelayInSeconds = DEFAULT_LONG_MESSAGES_DELAY_SECONDS;
    private static int smallAttachmentsDelayInMinutes = SMALL_ATTACHMENTS_DELAY_MINUTES;
    private static int bigAttachmentsDelayInMinutes = BIG_ATTACHMENTS_DELAY_MINUTES;
    private static int delayToSendMessagesInMillis = DEFAULT_DELAY_TO_SEND_MESSAGES_MILLISECONDS;
    private static String bigFileNamePath = DEFAULT_BIG_FILE_NAME_PATH;
    private static String smallFileNamePath = DEFAULT_SMALL_FILE_NAME_PATH;

    // ~~~~~~~~~~~~~~~~~~~ VARIABLES ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    /**
     * Server where the chat server runs, e.g. {@code mps001}
     */
    private static String server = null;

    /**
     * FQDN of the chat server.
     * {@code user001@domain, room001@conference.domain}.
     */
    private static String domain = null;

    /**
     * Port that the chat server is listening for messages.
     */
    private static int port = DEFAULT_PORT;

    /**
     * Password of all test user(s).
     */
    private static String password = DEFAULT_PASSWORD;

    /**
     * Full name of observer client, e.g. user050@<domain>/Spark-2.8.3
     */
    private static String observer;

    /**
     * Number of users simulated. (50 users in 5 rooms)
     */
    private static int numberOfUsers = DEFAULT_NUM_OF_USERS;

    /**
     * Number of chat rooms (roomXXX) (default 5 rooms).
     */
    private static int numberOfRooms = DEFAULT_NUM_OF_ROOMS;

    /**
     * @{code true} if test small attachments (file transfers).
     */
    private static boolean testAttachments;
    /**
     * @{code true} if test big attachments (file transfers).
     */
    private static boolean testBigAttachments;
    /**
     * @{code true} if test long messages.
     */
    private static boolean testLongMessages;

    /**
     * Simulated test users
     */
    private static final List<User> users = new ArrayList<>(numberOfUsers);

    /**
     * Chat rooms
     */
    private static final List<ChatRoom> chatRooms = new ArrayList<>(numberOfRooms);

    /**
     * User (e.g. {@code user001}). Functionality delegated to @{see
     * XmppManager}.
     */
    final class User {

        private final String username;
        private final String password;
        private final String domain;
        private final XmppManager xmppManager; // delegate to him
        private MultiUserChat joinedChatRoom;

        public User(String username, String password, String domain, XmppManager xmppManager) {
            this.username = username;
            this.password = password;
            this.domain = domain;
            this.xmppManager = xmppManager;
        }

        public String getUsername() {
            return username;
        }

        public String getPassword() {
            return password;
        }

        public String getJabberID() {
            return username + "@" + domain;
        }

        public void connect() {
            xmppManager.connect();
        }

        public void disconnect() {
            xmppManager.destroy();
            LOG.info("User " + username + " disconnected.");
        }

        public void login() {
            xmppManager.login(username, password);
        }

        public void setStatus(boolean available, String status) {
            xmppManager.setStatus(available, status);
        }

        public void sendMessage(String toJID, String message) {
            xmppManager.sendMessage(toJID, message);
        }

        public void receiveMessage() {
            xmppManager.receiveMessage();
        }

        public boolean sendAttachment(String toJID, String path) {
            return xmppManager.sendAttachment(toJID, "Smack", path);
        }

        public void receiveAttachment() {
            xmppManager.receiveAttachment(username);
        }

        public void joinChatRoom(String roomName, String nickname) {
            joinedChatRoom = xmppManager.joinChatRoom(roomName, nickname);
        }

        public void leaveChatRoom() {
            try {
                joinedChatRoom.leave();
            } catch (SmackException.NotConnectedException | InterruptedException ex) {
                LOG.severe(ex.getLocalizedMessage());
            }
        }

        public void sendMessageToChatRoom(String message) {
            xmppManager.sendMessageToChatRoom(joinedChatRoom, message);
        }

        public String getJoinedChatRoom() {
            return joinedChatRoom.getRoom().toString();
        }

        public void addRosterListener() {
            xmppManager.rosterChanged();
        }
    }

    /**
     * Chat room, e.g. {@code room001}
     */
    final class ChatRoom {

        private final String name;
        private final String domain;

        public ChatRoom(String name, String domain) {
            this.name = name;
            this.domain = domain;
        }

        public String getName() {
            return name + "@conference." + domain;
        }

    }

    /**
     * Transfer the file to all other users.
     *
     * @param i i-th user
     * @param path path of the file to be sent
     */
    private synchronized void transferFile(int i, String path) {
        int j;
        for (j = 1; j <= numberOfUsers; j++) {
            if (i != j) {
                try {
                    int delay = testBigAttachments ? bigAttachmentsDelayInMinutes : smallAttachmentsDelayInMinutes;
                    Thread.sleep(delay);
                    if (users.get(i - 1).sendAttachment(users.get(j - 1).getJabberID(), path)) {
                        LOG.info("Attachment " + path + " sent from " + users.get(i - 1).getJabberID() + " to " + users.get(j - 1).getJabberID());
                    } else {
                        LOG.severe("Attachment " + path + " from " + users.get(i - 1).getJabberID() + " to " + users.get(j - 1).getJabberID() + "  was not sent!");
                    }
                } catch (InterruptedException ie) {
                    Thread.currentThread().interrupt(); // reset the flag
                }
            }
        }
    }

    /**
     * Exchange file attachments between users.
     *
     * @param path path of the file to send
     * @see #transferFile(int, java.lang.String)
     */
    private void fileTransferTask(String path) {
        for (int i = 1; i <= numberOfUsers; i++) {
            transferFile(i, path);
        }
    }

    /**
     * Each user sends a message to a chat room.
     */
    private synchronized void chatRoomMessageTask() {
        for (int i = 1; i <= numberOfUsers; i++) {
            String message = testLongMessages ? LONG_MESSAGE : MESSAGE;
            User user = users.get(i - 1);
            try {
                Thread.sleep(delayToSendMessagesInMillis); // sleep 1"
                user.sendMessageToChatRoom(message);
                LOG.info(user.getJabberID() + " sent " + (testLongMessages ? "long" : "short") + " message to " + user.getJoinedChatRoom());
            } catch (InterruptedException ie) {
                Thread.currentThread().interrupt(); // reset the flag
            }
        }
    }

    /**
     * @param args the command line arguments
     */
    public static void main(String[] args) {
        parseCLIArguments(args);
        final XmppLoadTest loadXmppTest = new XmppLoadTest();
        loadProperties(PROPERTIES_FILE);
        // When used with ExecutorService it hangs the process and you need kill -9 to kill it
//        Runtime.getRuntime().addShutdownHook(new Thread() {
//            @Override
//            public void run() {
//                LOG.info("Shutting down ...");
//                // disconnect
//                for (User user : users) {
//                    user.leaveChatRoom();
//                    user.disconnect();
//                }
//            }
//        });

        init(loadXmppTest);
        performLoad(loadXmppTest);
    }

    /**
     * Initialise users and chat rooms.
     *
     * @param loadXmppTest
     */
    private static void init(XmppLoadTest loadXmppTest) {
        XmppManager xmppManager = new XmppManager(server, domain, port);
        for (int i = 1; i <= numberOfUsers; i++) {
            User user = loadXmppTest.new User("user" + String.format("%03d", i), password, domain, xmppManager);
            user.connect();
            user.login();
            user.setStatus(true, "Hello from " + user.getUsername());
            user.addRosterListener();
            users.add(user);
            if (testAttachments || testBigAttachments) {
                user.receiveAttachment();
            }
        }

        for (int i = 0; i < numberOfRooms; i++) {
            chatRooms.add(loadXmppTest.new ChatRoom("room" + String.format("%03d", i + 1), domain));
        }

        if (!testAttachments && !testBigAttachments) {
            // join chatrooms
            for (int i = 1; i <= numberOfUsers; i++) {
                ChatRoom chatRoom = chatRooms.get((i - 1) % numberOfRooms);
                User user = users.get(i - 1);
                user.joinChatRoom(chatRoom.getName(), user.getJabberID());
            }
        }
    }

    /**
     * Perform the load test.
     * 
     * @param loadXmppTest 
     */
    private static void performLoad(final XmppLoadTest loadXmppTest) {
        if (testAttachments || testBigAttachments) {  // send attachments
            String filePath = testBigAttachments ? bigFileNamePath : smallFileNamePath;
            int delay = testBigAttachments ? bigAttachmentsDelayInMinutes : smallAttachmentsDelayInMinutes;
            final Runnable task = () -> {
                while (true) {
                    if (Thread.currentThread().isInterrupted()) {
                        return;
                    }
                    loadXmppTest.fileTransferTask(filePath);
                }
            };
            scheduler.scheduleWithFixedDelay(task, 0, delay, MINUTES);
        } else { // send messages to chat rooms
            final Runnable task = () -> {
                while (true) {
                    if (Thread.currentThread().isInterrupted()) {
                        return;
                    }
                    loadXmppTest.chatRoomMessageTask();
                }
            };
            int every = testLongMessages ? longMessagesDelayInSeconds : shortMessagesDelayInSeconds;
            scheduler.scheduleWithFixedDelay(task, 0, every, SECONDS); // every x seconds
        }
        if (Thread.currentThread().isInterrupted()) {
            scheduler.shutdown();
        }
    }    
    
    /**
     * Read delays from the properties file. These could also be defined as
     * command line parameters.
     *
     * @param path of the properties file
     */
    private static void loadProperties(String path) {
        try (InputStream input = new FileInputStream(path)) {
            Properties prop = new Properties();
            prop.load(input);

            shortMessagesDelayInSeconds = Integer.parseInt(prop.getProperty("SHORT_MESSAGES_DELAY_SECONDS", String.valueOf(DEFAULT_SHORT_MESSAGES_DELAY_SECONDS)));
            longMessagesDelayInSeconds = Integer.parseInt(prop.getProperty("LONG_MESSAGES_DELAY_SECONDS", String.valueOf(DEFAULT_LONG_MESSAGES_DELAY_SECONDS)));
            smallAttachmentsDelayInMinutes = Integer.parseInt(prop.getProperty("SMALL_ATTACHMENTS_DELAY_MINUTES", String.valueOf(SMALL_ATTACHMENTS_DELAY_MINUTES)));
            bigAttachmentsDelayInMinutes = Integer.parseInt(prop.getProperty("BIG_ATTACHMENTS_DELAY_MINUTES", String.valueOf(BIG_ATTACHMENTS_DELAY_MINUTES)));
            delayToSendMessagesInMillis = Integer.parseInt(prop.getProperty("DELAY_TO_SEND_MESSAGES_MILLISECONDS", String.valueOf(DEFAULT_DELAY_TO_SEND_MESSAGES_MILLISECONDS)));
            smallFileNamePath = prop.getProperty("SMALL_FILE_NAME_PATH", DEFAULT_SMALL_FILE_NAME_PATH);
            bigFileNamePath = prop.getProperty("BIG_FILE_NAME_PATH", DEFAULT_BIG_FILE_NAME_PATH);
        } catch (IOException | NumberFormatException ex) {
            LOG.warning(ex.getMessage());
        }
    }

    /**
     * Parse command line arguments using Apache Commons CLI.
     *
     * @param args
     */
    private static void parseCLIArguments(String[] args) {
        // Definition stage
        Options options = new Options();

        options.addOption(Option.builder("s")
                .longOpt("server")
                .desc("Server")
                .hasArg()
                .argName("server")
                .required()
                .build());

        options.addOption(Option.builder("d")
                .longOpt("domain")
                .desc("Domain")
                .hasArg()
                .argName("domain")
                .required()
                .build());

        options.addOption(Option.builder("p")
                .longOpt("password")
                .desc("Password")
                .hasArg()
                .argName("password")
                .required()
                .build());

        options.addOption(Option.builder("n")
                .longOpt("number")
                .desc("Number of users")
                .hasArg()
                .argName("number")
                .required()
                .build());

        options.addOption(Option.builder("o")
                .longOpt("observer")
                .desc("Observer")
                .hasArg()
                .argName("obs")
                .build());

        options.addOption(Option.builder("a")
                .longOpt("attachment")
                .desc("Test attachments")
                .build());

        options.addOption(Option.builder("b")
                .longOpt("big")
                .desc("Test big attachments or messages")
                .build());

        // Parsing stage
        CommandLineParser parser = new DefaultParser();
        HelpFormatter formatter = new HelpFormatter();
        CommandLine cmd = null;

        try {
            cmd = parser.parse(options, args);
        } catch (ParseException e) {
            LOG.severe(e.getMessage());
            formatter.printHelp("java -Djava.util.logging.config.file=logging.properties -jar XmppLoadTest.jar", options);
            usage();
            System.exit(1);
        }
        // Interrogation stage
        if (cmd != null) {
            if (cmd.hasOption("a")) {  // send attachments
                if (cmd.hasOption("b")) {
                    testBigAttachments = true;
                } else {
                    testAttachments = true;
                }
            } else { // send messages
                if (cmd.hasOption("b")) {
                    testLongMessages = true;
                }
            }
            if (cmd.hasOption("p")) {
                password = cmd.getOptionValue("p");
                if (password == null) {
                    password = DEFAULT_PASSWORD;
                }
            }
            if (cmd.hasOption("d")) {
                domain = cmd.getOptionValue("d");
                if (domain == null || domain.isEmpty()) {
                    throw new NullPointerException("XMPP FQDN cannot be null or empty!");
                }
            }
            if (cmd.hasOption("s")) {
                server = cmd.getOptionValue("s");
                if (server == null || server.isEmpty()) {
                    if (domain == null || domain.isEmpty()) {
                        throw new NullPointerException("Server and XMPP FQDN cannot be null or empty!");
                    } else {
                        server = domain;
                        LOG.warning("Server was empty. " + domain + " was used.");
                    }
                }
            }
            if (cmd.hasOption("n")) {
                try {
                    numberOfUsers = Integer.valueOf(cmd.getOptionValue("n"));
                } catch (NumberFormatException nfe) {
                    LOG.warning("Wrong number of users argument: " + cmd.getOptionValue("n"));
                    numberOfUsers = DEFAULT_NUM_OF_USERS;
                }
                if (numberOfUsers <= 0 || numberOfUsers > MAX_NUM_OF_USERS_PER_SESSION) {
                    LOG.warning("Number of users must be in [1, " + MAX_NUM_OF_USERS_PER_SESSION + "] but was: " + cmd.getOptionValue("n"));
                    numberOfUsers = DEFAULT_NUM_OF_USERS;
                }
            }
            if (cmd.hasOption("o")) {
                observer = cmd.getOptionValue("o");
                if (observer == null || observer.isEmpty()) {
                    throw new NullPointerException("Observer cannot be null or empty!");
                }
            }
        }
    }

    /**
     * Extra usage message similar to {@code loadtest.c}.
     */
    private static void usage() {
        System.out.println("Usage : java -Djava.util.logging.config.file=logging.properties -jar XmppLoadTest.jar -s <server> -d <domain> -p <password> -n <numberOfUsers> [-o <observer_full_name>] [-a] [-b]\n"
                + "        jabber id : userXXX@<domain>\n"
                + "        chatroom  : roomXXX@conference.<domain>\n"
                + "        observer  : userXXX@<domain>/Spark-2.8.3<version> (just to test)\n"
                + "        10 users per chatroom\n"
                + "        5 chatrooms\n"
                + "        Use: \n"
                + "        -a to test small attachments (file transfers) or\n"
                + "        -a -b to test big attachments (file transfers)\n"
                + "        or: \n"
                + "        -b to test long messages\n");
    }
}
