package test.xmpp.xmpploadtest;

import org.jivesoftware.smack.*;
import org.jivesoftware.smack.ConnectionConfiguration.SecurityMode;
import org.jivesoftware.smack.chat2.Chat;
import org.jivesoftware.smack.chat2.ChatManager;
import org.jivesoftware.smack.java7.Java7SmackInitializer;
import org.jivesoftware.smack.packet.Message;
import org.jivesoftware.smack.packet.Presence;
import org.jivesoftware.smack.packet.Presence.Type;
import org.jivesoftware.smack.roster.Roster;
import org.jivesoftware.smack.roster.RosterEntry;
import org.jivesoftware.smack.roster.RosterListener;
import org.jivesoftware.smack.sasl.SASLMechanism;
import org.jivesoftware.smack.sasl.javax.SASLPlainMechanism;
import org.jivesoftware.smack.tcp.XMPPTCPConnection;
import org.jivesoftware.smack.tcp.XMPPTCPConnectionConfiguration;
import org.jivesoftware.smack.util.TLSUtils;
import org.jivesoftware.smackx.filetransfer.*;
import org.jivesoftware.smackx.muc.MultiUserChat;
import org.jivesoftware.smackx.muc.MultiUserChatException;
import org.jivesoftware.smackx.muc.MultiUserChatManager;
import org.jxmpp.jid.EntityBareJid;
import org.jxmpp.jid.Jid;
import org.jxmpp.jid.impl.JidCreate;
import org.jxmpp.jid.parts.Resourcepart;
import org.jxmpp.stringprep.XmppStringprepException;

import javax.net.ssl.SSLSession;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.nio.file.Paths;
import java.security.KeyManagementException;
import java.security.NoSuchAlgorithmException;
import java.util.Collection;
import java.util.logging.Logger;

public class XmppManager {

    private static final Logger LOG = Logger.getLogger(XmppManager.class.getName());
    private static final int PACKET_REPLY_TIMEOUT = 500; // millis
    private static final int BUFFER_SIZE = 1024;
    private final String server, domain;
    private final int port;
    private String resource = "Smack";
    private AbstractXMPPConnection connection;

    public XmppManager(String server, String domain, int port) {
        this.server = server;
        this.domain = domain;
        this.port = port;
    }

    public void connect() {
        LOG.info(String.format("Initializing connection to server %1$s port %2$d", domain, port));
        new Java7SmackInitializer().initialize();
        SmackConfiguration.DEBUG = true;

        SmackConfiguration.setDefaultReplyTimeout(PACKET_REPLY_TIMEOUT);

        XMPPTCPConnectionConfiguration.Builder builder = XMPPTCPConnectionConfiguration.builder();
        try {
            builder.setXmppDomain(JidCreate.domainBareFrom(domain))
                    .setHost(server)
                    .setPort(port)
                    .setResource(resource)
                    //                .setSocketFactory(new SSLSocketFactoryImpl())
                    .setSecurityMode(SecurityMode.disabled)
                    .setHostnameVerifier((String hostname, SSLSession session) -> true);
        } catch (XmppStringprepException ex) {
            LOG.severe(ex.getLocalizedMessage());
        }

//        SASLMechanism mechanism = new SASLDigestMD5Mechanism();
        SASLMechanism mechanism = new SASLPlainMechanism();
        SASLAuthentication.registerSASLMechanism(mechanism);
        SASLAuthentication.unBlacklistSASLMechanism("PLAIN");
        SASLAuthentication.blacklistSASLMechanism("SCRAM-SHA-1");
        SASLAuthentication.unBlacklistSASLMechanism("DIGEST-MD5");        
        
        try {
            builder = TLSUtils.acceptAllCertificates(builder);
        } catch (KeyManagementException | NoSuchAlgorithmException ex) {
            LOG.severe(ex.getLocalizedMessage());
        }
        XMPPTCPConnection.setUseStreamManagementDefault(true);
        XMPPTCPConnectionConfiguration config = builder.build();

        connection = new XMPPTCPConnection(config);
        connection.setReplyTimeout(1000L);
        try {
            connection.connect();
        } catch (SmackException | IOException | XMPPException | InterruptedException ex) {
            LOG.severe(ex.getLocalizedMessage());
        }

        LOG.info("Connected: " + connection.isConnected());
    }

    public void login(String username, String password) {
        if (connection != null && connection.isConnected()) {
            try {
                connection.login(username, password);
            } catch (XMPPException | SmackException | IOException | InterruptedException ex) {
                LOG.severe(ex.getLocalizedMessage());
            }
        }
        LOG.info(username + " authenticated? " + connection.isAuthenticated());
    }

    public void setStatus(boolean available, String status) {
        Presence.Type type = available ? Type.available : Type.unavailable;
        Presence presence = new Presence(type);

        presence.setStatus(status);
        try {
            connection.sendStanza(presence);
        } catch (SmackException.NotConnectedException | InterruptedException ex) {
            LOG.severe(ex.getLocalizedMessage());
        }
    }

    public void destroy() {
        if (connection != null && connection.isConnected()) {
            connection.disconnect();
        }
    }

    /**
     * Send message to another user.
     *
     * @param buddyJID recipient
     * @param message to send
     */
    public void sendMessage(String buddyJID, String message) {
        LOG.info(String.format("Sending message '%1$s' to user %2$s", message, buddyJID));
        try {
            Chat chat = ChatManager.getInstanceFor(connection).chatWith(JidCreate.entityBareFrom(buddyJID));
            chat.send(message);
        } catch (XmppStringprepException | SmackException.NotConnectedException | InterruptedException ex) {
            LOG.severe(ex.getLocalizedMessage());
        }
    }

    public void receiveMessage() {
        ChatManager.getInstanceFor(connection).addIncomingListener(
                (EntityBareJid from, Message message, Chat chat) -> {
                    LOG.info("New message from " + from + ": " + message.getBody());
                });
    }

    /**
     * File transfer.
     *
     * @param buddyJID recipient
     * @param res e.g. "Spark-2.8.3", default "Smack" (cannot be empty or
     * null)
     * @param path path of the file attachment to send
     * @return {@code true} if file transfer was successful
     */
    public boolean sendAttachment(String buddyJID, String res, String path) {
        LOG.info(String.format("Sending attachment '%1$s' to user %2$s", path, buddyJID));
        FileTransferManager fileTransferManager = FileTransferManager.getInstanceFor(connection);
        OutgoingFileTransfer fileTransfer = null;
        try {
            fileTransfer = fileTransferManager.createOutgoingFileTransfer(JidCreate.entityFullFrom(buddyJID + "/" + (res != null ? res : resource)));
        } catch (XmppStringprepException ex) {
            LOG.severe(ex.getLocalizedMessage());
            return false;
        }
        if (fileTransfer != null) {
            OutgoingFileTransfer.setResponseTimeout(15 * 60 * 1000);
            LOG.info("status is:" + fileTransfer.getStatus());
            FileTransferNegotiator.IBB_ONLY = true;
            File file = Paths.get(path).toFile();
            if (file.exists()) {
                try {
                    fileTransfer.sendFile(file, "sending attachment...");
                } catch (SmackException ex) {
                    LOG.severe(ex.getLocalizedMessage());
                    return false;
                }
                LOG.info("status is:" + fileTransfer.getStatus());
                if (hasError(fileTransfer)) {
                    LOG.severe(getErrorMessage(fileTransfer));
                    return false;
                } else {
                    return monitorFileTransfer(fileTransfer, buddyJID);
                }
            } else {
                try {
                    throw new FileNotFoundException("File " + path + " not found!");
                } catch (FileNotFoundException ex) {
                    LOG.severe(ex.getLocalizedMessage());
                    return false;
                }
            }
        }
        return true;
    }

    private boolean negotiationFailed(FileTransfer fileTransfer) {
        return fileTransfer.getException() != null && (fileTransfer.getStatus() == FileTransfer.Status.initial
                || fileTransfer.getStatus() == FileTransfer.Status.negotiated
                || fileTransfer.getStatus() == FileTransfer.Status.negotiating_stream
                || fileTransfer.getStatus() == FileTransfer.Status.negotiating_transfer);
    }

    private boolean isComplete(FileTransfer fileTransfer) {
        return (fileTransfer.getStatus() == FileTransfer.Status.complete && fileTransfer.getAmountWritten() == fileTransfer.getFileSize());
    }

    private boolean inProgress(FileTransfer fileTransfer) {
        return (fileTransfer.getStatus() == FileTransfer.Status.in_progress);
    }

    private boolean isCancelled(FileTransfer fileTransfer) {
        return (fileTransfer.getStatus() == FileTransfer.Status.complete);
    }

    private boolean isRejected(FileTransfer fileTransfer) {
        return (fileTransfer.getStatus() == FileTransfer.Status.refused);
    }

    private boolean hasError(FileTransfer fileTransfer) {
        return (fileTransfer.getStatus() == FileTransfer.Status.error);
    }

    private String getErrorMessage(FileTransfer fileTransfer) {
        if (fileTransfer.getStatus() == FileTransfer.Status.error && fileTransfer.getError() != null) {
            return fileTransfer.getError().getMessage();
        } else if (fileTransfer.getException() != null) {
            if (fileTransfer.getStatus() == FileTransfer.Status.negotiating_transfer) {
                if (fileTransfer.getException() instanceof SmackException.NoResponseException) {
                    return "The recipient did not responded within the reply timeout.";
                }
            }
            return fileTransfer.getException().getMessage();
        } else {
            return fileTransfer.getStatus().toString();
        }
    }

    /**
     * Monitor file transfer.
     *
     * @param fileTransfer
     * @param buddyJID
     * @return {@code false} if file transfer failed.
     */
    private boolean monitorFileTransfer(FileTransfer fileTransfer, String buddyJID) {
        while (!fileTransfer.isDone()) {
            if (isRejected(fileTransfer) || isCancelled(fileTransfer)
                    || negotiationFailed(fileTransfer) || hasError(fileTransfer)) {
                LOG.severe("Could not send/receive the file to/from " + buddyJID + "." + fileTransfer.getError());
                LOG.severe(getErrorMessage(fileTransfer));
                return false;
            } else if (inProgress(fileTransfer)) {
                LOG.info("File transfer status: " + fileTransfer.getStatus() + ", progress: " + fileTransfer.getProgress());
            }
            try {
                Thread.sleep(1000);
            } catch (InterruptedException ex) {
                LOG.severe(ex.getLocalizedMessage());
            }
        }
        if (isComplete(fileTransfer)) {
            LOG.info(fileTransfer.getFileName() + " has been successfully transferred.");
            LOG.info("The file transfer is " + (fileTransfer.isDone() ? "done." : "not done."));
            return true;
        }
        return true;
    }

    public void receiveAttachment(String username) {
        final FileTransferManager manager = FileTransferManager.getInstanceFor(connection);
        manager.addFileTransferListener((FileTransferRequest request) -> {
            // Check to see if the request should be accepted
            if (request.getFileName() != null) {
                StringBuilder sb = new StringBuilder(BUFFER_SIZE);
                try {
                    // Accept it
                    IncomingFileTransfer transfer = request.accept();
                    String filename = transfer.getFileName() + "_" + username;
                    transfer.receiveFile(new File(filename));
                    while (!transfer.isDone()) {
                        try {
                            Thread.sleep(1000);
                            LOG.info("STATUS: " + transfer.getStatus()
                                    + " SIZE: " + sb.toString().length()
                                    + " Stream ID : " + transfer.getStreamID());
                        } catch (Exception e) {
                            LOG.severe(e.getMessage());
                        }
                        if (transfer.getStatus().equals(FileTransfer.Status.error)) {
                            LOG.severe(transfer.getStatus().name());
                        }
                        if (transfer.getException() != null) {
                            LOG.severe(transfer.getException().getLocalizedMessage());
                        }
                    }
                    LOG.info("File received " + request.getFileName());
                } catch (SmackException | IOException ex) {
                    LOG.severe(ex.getLocalizedMessage());
                }
            } else {
                try {
                    // Reject it
                    request.reject();
                    LOG.warning("File rejected " + request.getFileName());
                } catch (SmackException.NotConnectedException | InterruptedException ex) {
                    LOG.severe(ex.getLocalizedMessage());
                }
            }
        });
    }

    public Roster createRosterFor(String user, String name) throws Exception {
        LOG.info(String.format("Creating roster for buddy '%1$s' with name %2$s", user, name));
        Roster roster = Roster.getInstanceFor(connection);
        roster.createEntry(JidCreate.bareFrom(user), name, null);
        return roster;
    }

    public void printRosters() throws Exception {
        Roster roster = Roster.getInstanceFor(connection);
        Collection<RosterEntry> entries = roster.getEntries();
        for (RosterEntry entry : entries) {
            LOG.info(String.format("Buddy: %s", entry.getName()));
        }
    }

    public void rosterChanged() {
        Roster roster = Roster.getInstanceFor(connection);
        roster.addRosterListener(new RosterListener() {

            @Override
            public void presenceChanged(Presence presence) {
                LOG.info("Presence changed: " + presence.getFrom() + " " + presence);
                resource = presence.getFrom().getResourceOrEmpty().toString();
            }

            @Override
            public void entriesAdded(Collection<Jid> clctn) {
            }

            @Override
            public void entriesUpdated(Collection<Jid> clctn) {
            }

            @Override
            public void entriesDeleted(Collection<Jid> clctn) {
            }
        });
    }

    public MultiUserChat joinChatRoom(String roomName, String nick) {
        try {
            MultiUserChatManager manager = MultiUserChatManager.getInstanceFor(connection);
            MultiUserChat muc = manager.getMultiUserChat(JidCreate.entityBareFrom(roomName));
            Resourcepart nickname = Resourcepart.from(nick);
            muc.join(nickname);
            LOG.info(muc.getNickname() + " joined chat room " + muc.getRoom());
            return muc;
        } catch (XmppStringprepException | SmackException.NotConnectedException | InterruptedException | SmackException.NoResponseException | XMPPException.XMPPErrorException | MultiUserChatException.NotAMucServiceException ex) {
            LOG.severe(ex.getLocalizedMessage());
        }
        return null;
    }

    public void sendMessageToChatRoom(MultiUserChat muc, String message) {
        try {
            muc.sendMessage(message);
            LOG.fine("Message '" + message + "' was sent to room '" + muc.getRoom() + "' by '" + muc.getNickname() + "'");
        } catch (InterruptedException | SmackException.NotConnectedException ex) {
            LOG.severe(ex.getLocalizedMessage());
        }
    }

}
